from pyspedas.projects.mms.mms_load_data import mms_load_data
from pyspedas.projects.mms.feeps_tools.mms_feeps_correct_energies import mms_feeps_correct_energies
from pyspedas.projects.mms.feeps_tools.mms_feeps_flat_field_corrections import mms_feeps_flat_field_corrections
from pyspedas.projects.mms.feeps_tools.mms_feeps_active_eyes import mms_feeps_active_eyes
from pyspedas.projects.mms.feeps_tools.mms_feeps_split_integral_ch import mms_feeps_split_integral_ch
from pyspedas.projects.mms.feeps_tools.mms_feeps_remove_bad_data import mms_feeps_remove_bad_data
from pyspedas.projects.mms.feeps_tools.mms_feeps_remove_sun import mms_feeps_remove_sun
from pyspedas.projects.mms.feeps_tools.mms_feeps_omni import mms_feeps_omni
from pyspedas.projects.mms.feeps_tools.mms_feeps_spin_avg import mms_feeps_spin_avg
from pyspedas.projects.mms.mms_config import CONFIG
from pyspedas.tplot_tools import time_clip as tclip

def recvary_log_filter(log):
    if 'record-varying' in log.msg:
        return False
    else:
        return True

def mms_load_feeps(trange=['2015-10-16', '2015-10-17'],
                   probe='1',
                   data_rate='srvy',
                   level='l2',
                   datatype='electron',
                   varformat=None,
                   varnames=[],
                   get_support_data=True,
                   suffix='',
                   time_clip=False,
                   no_update=False,
                   available=False,
                   notplot=False,
                   no_flatfield_corrections=False,
                   data_units=['count_rate', 'intensity'],
                   latest_version=False,
                   major_version=False,
                   min_version=None,
                   cdf_version=None,
                   spdf=False,
                   always_prompt=False,
                   filter_recvary_warnings=True,
                   quality_flag=3,
                   get_err=False,
                   ):
    """
    Load data from the MMS Fly's Eye Energetic Particle Sensor (FEEPS)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [start time, end time] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2015-10-16','2015-10-17']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4'].
            Default: '1'

        data_rate : str or list of str
            instrument data rates for FEEPS include ['brst', 'srvy'].
            Default: 'srvy'

        level : str
            indicates level of data processing. the default if no level is specified is 'l2'
            Default: 'l2'

        datatype : str or list of str
            Valid datatypes for FEEPS are::

                       L2, L1b: ['electron', 'ion']
                       L1a: ['electron-bottom', 'electron-top', 'ion-bottom', 'ion-top']

            Default: 'electron'

        data_units : str or list of str
            Unit types to be loaded, options are 'count_rate', 'intensity'
            Default: ['count_rate', 'intensity']

        no_flatfield_corrections: bool
            If True, no flatfield corrections are performed.
            Default: False

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: True

        time_clip: bool
            Data will be clipped to the exact trange specified by the trange keyword.
            Default: False
            
        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None (all variables are loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded.
            Default: []

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.
            Default: None

        notplot: bool
            If True, then data are returned in a hash table instead of 
            being stored in tplot variables (useful for debugging, and
            access to multidimensional data products)
            Default: False

        available: bool
            If True, simply return the available data files (without downloading)
            for the requested parameters
            Default: False

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten
            Default: False

        cdf_version: str
            Specify a specific CDF version # to load (e.g., cdf_version='4.3.0')
            Default: None

        min_version: str
            Specify a minimum CDF version # to load
            Default: None

        latest_version: bool
            Only grab the latest CDF version in the requested time interval

        major_version: bool
            Only open the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval

        always_prompt: bool
            Set this keyword to always prompt for the user's username and password;
            useful if you accidentally save an incorrect password, or if your SDC password has changed

        spdf: bool
            If True, download the data from the SPDF instead of the SDC

        filter_recvary_warnings: bool
            If True, capture warnings from cdf_to_tplot and filter out the ones complaining about non-record-varying support data with timestamps

        quality_flag: int
            Value can range from 2-4. Determines which eyes to use to calculate omnidirectional flux at a given time
            based on how many sectors were masked for sunlight. For BRST or RAW::

                Green (Quality Indicator = 0): No contaminated or masked spin sectors (best data quality)
                Red (Quality Indicator = 3): Contaminated spin sector is associated with BURST and RAW data
                    (this data is not recommended for scientific use)
                Grey (Quality Indicator = 4): FEEPS calibration data (this data is generated by the instrument itself
                    and is not recommended for scientific use)

            For data in SURVEY mode, there are 5 possible quality indicators::

                Green (Quality Indicator = 0): No contaminated or masked spin sectors (best data quality)
                Yellow (Quality Indicator = 1): Among the spin sectors used to create a SURVEY aggregate
                     spin sector, onboard masking is successfully applied to a minimum of 1 and a maximum of 7
                     spin sectors and there is no contamination in any of the individual spin sectors. Note that
                     since this indicator applies only when onboard masking is applied, some sectors are not
                     observed.
                Orange (Quality Indicator = 2): There is contamination of between 0 and 50% of the notmasked sectors. Warning: this data contains contamination.
                Red (Quality Indicator = 3): Same as orange except that there is contamination of 50% or
                     more of the not-masked sectors. IN ADDITION, this indicator applies to the case where
                     onboard masking is applied to all 8 spin sectors in the aggregate sector.
                Grey (Quality Indicator = 4): FEEPS calibration data (this data is generated by the instrument
                     itself and is not recommended for scientific use)

        get_err: bool
            Gets the error in the omnidirectional l2 intensity.  Default: False. User MUST run mms_load_feeps for the
            l1b counts data FIRST to use this function. (Note: the FEEPS l1b data needed for this switch is restricted
            to MMS team members at present.)


    Returns
    --------
        list of str
            List of tplot variables created.

    Example
    -------

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> feeps_data = pyspedas.projects.mms.mms_load_feeps(trange=['2015-10-16', '2015-10-17'], probe='1', datatype='electron')
    >>> tplot(['mms1_epd_feeps_srvy_l2_electron_intensity_omni_spin', 'mms1_epd_feeps_srvy_l2_electron_intensity_omni'])



    """

    # For log filtering
    from pyspedas import logger
    # as of 3 July 2023, there's a mixture of v7.x.x and v6.x.x files at the SDC
    # these files aren't compatible, so we need to only load the latest major version
    # to avoid crashes (unless otherwise specified)
    if not latest_version and not major_version and min_version is None and cdf_version is None:
        major_version = True

    if filter_recvary_warnings:
        logger.addFilter(recvary_log_filter)

    tvars = mms_load_data(trange=trange, notplot=notplot, probe=probe, data_rate=data_rate, level=level, instrument='feeps',
            datatype=datatype, varformat=varformat, varnames=varnames, get_support_data=get_support_data, suffix=suffix,
            no_update=no_update, available=available, latest_version=latest_version,
            major_version=major_version, min_version=min_version, cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)

    if filter_recvary_warnings:
        logger.removeFilter(recvary_log_filter)

    if tvars == [] or available or notplot or CONFIG['download_only'] or tvars is None:
        return tvars

    probes = probe if isinstance(probe, list) else [probe]
    data_rates = data_rate if isinstance(data_rate, list) else [data_rate]
    levels = level if isinstance(level, list) else [level]
    datatypes = datatype if isinstance(datatype, list) else [datatype]
    data_units = data_units if isinstance(data_units, list) else [data_units]

    probes = [str(p) for p in probes]

    mms_feeps_correct_energies(probes, data_rate, level=level, suffix=suffix)

    if not no_flatfield_corrections:
        mms_feeps_flat_field_corrections(probes=probes, data_rate=data_rate, suffix=suffix)

    for probe in probes:
        for lvl in levels:
            for drate in data_rates:
                for datatype in datatypes:
                    mms_feeps_remove_bad_data(trange=trange, probe=probe, data_rate=drate, datatype =datatype, level=lvl, suffix=suffix)

                    for data_unit in data_units:
                        eyes = mms_feeps_active_eyes(trange, probe, drate, datatype, lvl)

                        split_vars = mms_feeps_split_integral_ch(data_unit, datatype, probe, suffix=suffix, data_rate=drate, level=lvl, sensor_eyes=eyes)

                        sun_removed_vars = mms_feeps_remove_sun(eyes, trange, probe=probe, datatype=datatype, data_units=data_unit, data_rate=drate, level=lvl, suffix=suffix)

                        omni_vars = mms_feeps_omni(eyes, probe=probe, datatype=datatype, data_units=data_unit, data_rate=drate, level=lvl, suffix=suffix, quality_flag=quality_flag, get_err=get_err)

                        if split_vars is not None:
                            tvars = tvars + split_vars

                        if sun_removed_vars is not None:
                            tvars = tvars + sun_removed_vars

                        if omni_vars is not None:
                            tvars = tvars + omni_vars

                        spin_avg_vars = mms_feeps_spin_avg(probe=probe, data_units=data_unit, datatype=datatype, data_rate=drate, level=lvl, suffix=suffix)

                        if spin_avg_vars is not None:
                            tvars.append(spin_avg_vars)

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import os
import time
import requests
import unittest
import subprocess

from pyspedas.tplot_tools import data_exists
from pyspedas.projects.cluster import load_csa, config

#==========================================================

# moto server mock details
localhost = "http://localhost:3000"
bucket_name = "test-bucket"

# Set up mock AWS environment variables (fake credentials)
os.environ["AWS_ACCESS_KEY_ID"] = "test"
os.environ["AWS_SECRET_ACCESS_KEY"] = "test"
os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

# Set environment to use the local Moto S3 server
# S3 ENDPOINT for fsspec
# ENDPOINT URL for cdflib/boto3
os.environ["AWS_S3_ENDPOINT"] = localhost
os.environ["AWS_ENDPOINT_URL"] = localhost

class LoadTestCases(unittest.TestCase):
    """
    Cloud Awareness Unit Tests

    Depends upon moto[server] package. Install via:
        pip install moto[server]

    These tests essentially create a local mock-AWS server as a background
    process at port 3000.
    Note: The environment variables are used as mock credentials in order
          to avoid having to pass the endpoint url to fsspec calls.
    """

    @classmethod
    def setUpClass(cls):
        # Start the moto server for S3 in the background
        # https://github.com/getmoto/moto/issues/4418
        cls.moto_server = subprocess.Popen(
                ["moto_server", "-p3000"],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        # Allow the server to start properly
        time.sleep(2)

        # Create a bucket using direct HTTP requests
        response = requests.put(f"http://localhost:3000/{bucket_name}")
        assert response.status_code == 200, "Bucket creation failed"

    @classmethod
    def tearDownClass(cls):
        # Terminate the moto server after tests
        cls.moto_server.terminate()
        cls.moto_server.communicate()
    
    def clean_data(self):
        # reset moto server to original state
        response = requests.post("http://localhost:3000/moto-api/reset")
        assert response.status_code == 200, "Moto Server reset failed"
        
        # create bucket again
        response = requests.put(f"http://localhost:3000/{bucket_name}")
        assert response.status_code == 200, "Bucket creation failed"

    #==========================================================================
    # Cloud Awareness Note: Cluster implementation does not stream data due to
    #                       lack of functionality already present in PySPEDAS.

    # Adapted unit tests for AWS-specific URI testing.
    def test_load_csa_CE_WBD_WAVEFORM_CDF_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2012-11-06T02:19:00Z', '2012-11-06T02:19:59Z']
        wbd_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CE_WBD_WAVEFORM_CDF'], time_clip=True)
        print(wbd_data)
        self.assertTrue('WBD_Elec' in wbd_data)
        self.assertTrue(data_exists('WBD_Elec'))
    
    def test_load_csa_CP_AUX_POSGSE_1M_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        pos_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_AUX_POSGSE_1M'], time_clip=True)
        self.assertTrue('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M' in pos_data)
        self.assertTrue(data_exists('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M'))
    
    def test_load_csa_CP_AUX_POSGSE_1M_data_prefix_none(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        pos_data = load_csa(probes=['C1'],
                            trange=trange,
                            prefix=None,
                            datatypes=['CP_AUX_POSGSE_1M'], time_clip=True)
        self.assertTrue('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M' in pos_data)
        self.assertTrue(data_exists('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M'))
    
    def test_load_csa_CP_AUX_POSGSE_1M_data_suffix_none(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        pos_data = load_csa(probes=['C1'],
                            trange=trange,
                            suffix=None,
                            datatypes=['CP_AUX_POSGSE_1M'], time_clip=True)
        self.assertTrue('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M' in pos_data)
        self.assertTrue(data_exists('sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M'))
    
    def test_load_csa_CP_AUX_POSGSE_1M_data_prefix_suffix(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        pos_data = load_csa(probes=['C1'],
                            trange=trange,
                            prefix='pre_',
                            suffix='_suf',
                            datatypes=['CP_AUX_POSGSE_1M'], time_clip=True)
        self.assertTrue('pre_sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M_suf' in pos_data)
        self.assertTrue(data_exists('pre_sc_r_xyz_gse__C1_CP_AUX_POSGSE_1M_suf'))
    
    def test_load_csa_CP_CIS_CODIF_HS_H1_MOMENTS_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_CIS-CODIF_HS_H1_MOMENTS'], time_clip=True)
        print(mom_data)
        self.assertTrue('density__C1_CP_CIS_CODIF_HS_H1_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_CODIF_HS_H1_MOMENTS'))

    def test_load_csa_CP_CIS_CODIF_HS_He1_MOMENTS_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_CIS-CODIF_HS_He1_MOMENTS'], time_clip=True)
        print(mom_data)
        self.assertTrue('density__C1_CP_CIS_CODIF_HS_He1_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_CODIF_HS_He1_MOMENTS'))

    def test_load_csa_CP_CIS_CODIF_HS_O1_MOMENTS_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_CIS-CODIF_HS_O1_MOMENTS'], time_clip=True)
        print(mom_data)
        self.assertTrue('density__C1_CP_CIS_CODIF_HS_O1_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_CODIF_HS_O1_MOMENTS'))

    def test_load_csa_CP_CIS_CODIF_PAD_HS_H1_PF_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_CIS-CODIF_PAD_HS_H1_PF'], time_clip=True)
        print(mom_data)
        self.assertTrue('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_H1_PF' in mom_data)
        self.assertTrue(data_exists('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_H1_PF'))

    def test_load_csa_CP_CIS_CODIF_PAD_HS_He1_PF_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_CIS-CODIF_PAD_HS_He1_PF'], time_clip=True)
        print(mom_data)
        self.assertTrue('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_He1_PF' in mom_data)
        self.assertTrue(data_exists('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_He1_PF'))

    def test_load_csa_CP_CIS_CODIF_PAD_HS_O1_PF_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_CIS-CODIF_PAD_HS_O1_PF'], time_clip=True)
        print(mom_data)
        self.assertTrue('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_O1_PF' in mom_data)
        self.assertTrue(data_exists('Differential_Particle_Flux__C1_CP_CIS_CODIF_PAD_HS_O1_PF'))

    def test_load_csa_CP_CIS_HIA_ONBOARD_MOMENTS_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_CIS-HIA_ONBOARD_MOMENTS'], time_clip=True)
        print(mom_data)
        self.assertTrue('density__C1_CP_CIS_HIA_ONBOARD_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_HIA_ONBOARD_MOMENTS'))

    # This test uses a different time range from test_load_csa_CP_CIS_HIA_ONBOARD_MOMENTS_data,
    # and loads for all four probes.
    def test_load_csa_mom_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        mom_data = load_csa(probes=['C1', 'C2', 'C3', 'C4'],
                            trange=['2003-08-17/16:40', '2003-08-17/16:45'],
                            datatypes=['CP_CIS-HIA_ONBOARD_MOMENTS'], time_clip=True)
        self.assertTrue('density__C1_CP_CIS_HIA_ONBOARD_MOMENTS' in mom_data)
        self.assertTrue(data_exists('density__C1_CP_CIS_HIA_ONBOARD_MOMENTS'))

    def test_load_csa_CP_CIS_HIA_PAD_HS_MAG_IONS_PF_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        mom_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_CIS-HIA_PAD_HS_MAG_IONS_PF'], time_clip=True)
        print(mom_data)
        self.assertTrue('Differential_Particle_Flux__C1_CP_CIS_HIA_PAD_HS_MAG_IONS_PF' in mom_data)
        self.assertTrue(data_exists('Differential_Particle_Flux__C1_CP_CIS_HIA_PAD_HS_MAG_IONS_PF'))

    def test_load_csa_CP_EDI_AEDC_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2005-08-01T00:00:00Z', '2005-08-02T00:00:00Z']
        edi_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_EDI_AEDC'], time_clip=True)
        print(edi_data)
        self.assertTrue('counts_GDU1_PA_90__C1_CP_EDI_AEDC' in edi_data)
        self.assertTrue(data_exists('counts_GDU1_PA_90__C1_CP_EDI_AEDC'))

    def test_load_csa_CP_EDI_MP_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        edi_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_EDI_MP'], time_clip=True)
        self.assertTrue('V_ed_xyz_gse__C1_CP_EDI_MP' in edi_data)
        self.assertTrue(data_exists('V_ed_xyz_gse__C1_CP_EDI_MP'))

    def test_load_csa_CP_EDI_SPIN_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        edi_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_EDI_SPIN'], time_clip=True)
        self.assertTrue('V_ed_xyz_gse__C1_CP_EDI_SPIN' in edi_data)
        self.assertTrue(data_exists('V_ed_xyz_gse__C1_CP_EDI_SPIN'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_EFW_L2_E3D_INERT_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_EFW_L2_E3D_INERT'], time_clip=True)
        self.assertTrue('E_Vec_xyz_ISR2__C1_CP_EFW_L2_E3D_INERT' in efw_data)
        self.assertTrue(data_exists('E_Vec_xyz_ISR2__C1_CP_EFW_L2_E3D_INERT'))

    def test_load_csa_CP_EFW_L2_P_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-04T00:00:00Z']
        efw_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_EFW_L2_P'], time_clip=True)
        self.assertTrue('Spacecraft_potential__C1_CP_EFW_L2_P' in efw_data)
        self.assertTrue(data_exists('Spacecraft_potential__C1_CP_EFW_L2_P'))

    def test_load_csa_CP_EFW_L2_V3D_INERT_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_EFW_L2_V3D_INERT'], time_clip=True)
        self.assertTrue('v_drift_ISR2__C1_CP_EFW_L2_V3D_INERT' in efw_data)
        self.assertTrue(data_exists('v_drift_ISR2__C1_CP_EFW_L2_V3D_INERT'))

    def test_load_csa_CP_EFW_L3_E3D_INERT_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_EFW_L3_E3D_INERT'], time_clip=True)
        self.assertTrue('E_Vec_xyz_ISR2__C1_CP_EFW_L3_E3D_INERT' in efw_data)
        self.assertTrue(data_exists('E_Vec_xyz_ISR2__C1_CP_EFW_L3_E3D_INERT'))

    def test_load_csa_CP_EFW_L3_P_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_EFW_L3_P'], time_clip=True)
        self.assertTrue('Spacecraft_potential__C1_CP_EFW_L3_P' in efw_data)
        self.assertTrue(data_exists('Spacecraft_potential__C1_CP_EFW_L3_P'))

    def test_load_csa_CP_EFW_L3_V3D_INERT_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        efw_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_EFW_L3_V3D_INERT'], time_clip=True)
        self.assertTrue('v_drift_ISR2__C1_CP_EFW_L3_V3D_INERT' in efw_data)
        self.assertTrue(data_exists('v_drift_ISR2__C1_CP_EFW_L3_V3D_INERT'))

    def test_load_csa_CP_FGM_5VPS_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        fgm_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_FGM_5VPS'], time_clip=True)
        self.assertTrue('B_vec_xyz_gse__C1_CP_FGM_5VPS' in fgm_data)
        self.assertTrue(data_exists('B_vec_xyz_gse__C1_CP_FGM_5VPS'))

    def test_load_csa_CP_FGM_FULL_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        fgm_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_FGM_FULL'], time_clip=True)
        self.assertTrue('B_vec_xyz_gse__C1_CP_FGM_FULL' in fgm_data)
        self.assertTrue(data_exists('B_vec_xyz_gse__C1_CP_FGM_FULL'))

    def test_load_csa_CP_FGM_SPIN_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        fgm_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_FGM_SPIN'], time_clip=True)
        self.assertTrue('B_vec_xyz_gse__C1_CP_FGM_SPIN' in fgm_data)
        self.assertTrue(data_exists('B_vec_xyz_gse__C1_CP_FGM_SPIN'))

    def test_load_csa_CP_PEA_MOMENTS_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        mom_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_PEA_MOMENTS'], time_clip=True)
        self.assertTrue('Data_Density__C1_CP_PEA_MOMENTS' in mom_data)
        self.assertTrue(data_exists('Data_Density__C1_CP_PEA_MOMENTS'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_PEA_PITCH_SPIN_DEFlux_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        pea_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_PEA_PITCH_SPIN_DEFlux'], time_clip=True)
        self.assertTrue('Data__C1_CP_PEA_PITCH_SPIN_DEFlux' in pea_data)
        self.assertTrue(data_exists('Data__C1_CP_PEA_PITCH_SPIN_DEFlux'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_PEA_PITCH_SPIN_DPFlux_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        pea_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_PEA_PITCH_SPIN_DPFlux'], time_clip=True)
        self.assertTrue('Data__C1_CP_PEA_PITCH_SPIN_DPFlux' in pea_data)
        self.assertTrue(data_exists('Data__C1_CP_PEA_PITCH_SPIN_DPFlux'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_PEA_PITCH_SPIN_PSD_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        pea_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_PEA_PITCH_SPIN_PSD'], time_clip=True)
        self.assertTrue('Data__C1_CP_PEA_PITCH_SPIN_PSD' in pea_data)
        self.assertTrue(data_exists('Data__C1_CP_PEA_PITCH_SPIN_PSD'))

    # compressed file error
    @unittest.skip
    def test_load_csa_CP_RAP_ESPCT6_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2008-02-03T00:00:00Z', '2008-02-05T00:00:00Z']
        rap_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_RAP_ESPCT6'], time_clip=True)
        print(rap_data)
        self.assertTrue('Electron_Dif_flux__C1_CP_RAP_ESPCT6' in rap_data)
        self.assertTrue(data_exists('Electron_Dif_flux__C1_CP_RAP_ESPCT6'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_ESPCT6_R_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_RAP_ESPCT6_R'], time_clip=True)
        self.assertTrue('Electron_Rate__C1_CP_RAP_ESPCT6_R' in rap_data)
        self.assertTrue(data_exists('Electron_Rate__C1_CP_RAP_ESPCT6_R'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_HSPCT_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_RAP_HSPCT'], time_clip=True)
        self.assertTrue('Proton_Dif_flux__C1_CP_RAP_HSPCT' in rap_data)
        self.assertTrue(data_exists('Proton_Dif_flux__C1_CP_RAP_HSPCT'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_HSPCT_R_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_RAP_HSPCT_R'], time_clip=True)
        self.assertTrue('Proton_Rate__C1_CP_RAP_HSPCT_R' in rap_data)
        self.assertTrue(data_exists('Proton_Rate__C1_CP_RAP_HSPCT_R'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_ISPCT_CNO_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_RAP_ISPCT_CNO'], time_clip=True)
        self.assertTrue('CNO_Dif_flux__C1_CP_RAP_ISPCT_CNO' in rap_data)
        self.assertTrue(data_exists('CNO_Dif_flux__C1_CP_RAP_ISPCT_CNO'))

    # multidimensional DEPEND_0 array for dsettings/caveat variables
    def test_load_csa_CP_RAP_ISPCT_He_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        rap_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_RAP_ISPCT_He'], time_clip=True)
        self.assertTrue('Helium_Dif_flux__C1_CP_RAP_ISPCT_He' in rap_data)
        self.assertTrue(data_exists('Helium_Dif_flux__C1_CP_RAP_ISPCT_He'))

    # This date returns data, but it's large enough (tens of megabytes that there's a significant
    # chance of the server not completing the request.
    @unittest.skip
    def test_load_csa_CP_STA_CS_HBR_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-01-31T00:00:00Z', '2001-01-31T23:59:59Z']
        sta_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_STA_CS_HBR'], time_clip=True)
        print(sta_data)
        self.assertTrue('Complex_Spectrum__C1_CP_STA_CS_HBR' in sta_data)
        self.assertTrue(data_exists('Complex_Spectrum__C1_CP_STA_CS_HBR'))

    # This date returns data, but it's large enough (tens of megabytes that there's a significant
    # chance of the server not completing the request.
    @unittest.skip
    def test_load_csa_CP_STA_CS_NBR_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-01-31T00:00:00Z', '2001-01-31T23:59:59Z']
        sta_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_STA_CS_NBR'], time_clip=True)
        print(sta_data)
        self.assertTrue('Complex_Spectrum__C1_CP_STA_CS_NBR' in sta_data)
        self.assertTrue(data_exists('Complex_Spectrum__C1_CP_STA_CS_NBR'))

    # compressed file error
    @unittest.skip
    def test_load_csa_CP_STA_CWF_GSE_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        sta_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_STA_CWF_GSE'], time_clip=True)
        self.assertTrue('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE' in sta_data)
        self.assertTrue(data_exists('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE'))

    # 404 error, no data?
    @unittest.skip
    def test_load_csa_CP_STA_CWF_HBR_ISR2_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        sta_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_STA_CWF_HBR_ISR2'], time_clip=True)
        print(sta_data)
        self.assertTrue('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE' in sta_data)
        self.assertTrue(data_exists('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE'))

    # 404 error, no data?
    @unittest.skip
    def test_load_csa_CP_STA_CWF_NBR_ISR2_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        sta_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_STA_CWF_NBR_ISR2'], time_clip=True)
        print(sta_data)
        self.assertTrue('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE' in sta_data)
        self.assertTrue(data_exists('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE'))

    # Spectrogram variables have string-valued DEPEND_2, which is not ISTP compliant.
    @unittest.skip
    def test_load_csa_CP_STA_PSD_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        sta_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_STA_PSD'], time_clip=True)
        print(sta_data)
        #self.assertTrue('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE' in sta_data)
        #self.assertTrue(data_exists('B_vec_xyz_Instrument__C1_CP_STA_CWF_GSE'))

    # data load seemed rather large; potential timeout issue (works separately)
    @unittest.skip
    def test_load_csa_CP_WBD_WAVEFORM_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-04T13:40:00Z', '2001-02-04T13:49:59Z']
        wbd_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_WBD_WAVEFORM'], time_clip=True)
        print(wbd_data)
        self.assertTrue('E__C1_CP_WBD_WAVEFORM' in wbd_data)
        self.assertTrue(data_exists('E__C1_CP_WBD_WAVEFORM'))

    def test_load_csa_CP_WHI_ELECTRON_DENSITY_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2018-01-12T00:00:00Z', '2018-01-13T00:00:00Z']
        whi_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_WHI_ELECTRON_DENSITY'], time_clip=True)
        print(whi_data)
        self.assertTrue('Electron_Density__C1_CP_WHI_ELECTRON_DENSITY' in whi_data)
        self.assertTrue(data_exists('Electron_Density__C1_CP_WHI_ELECTRON_DENSITY'))

    def test_load_csa_CP_CP_WHI_NATURAL_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2001-02-01T00:00:00Z', '2001-02-02T00:00:00Z']
        whi_data = load_csa(probes=['C1'],
                            trange=trange,
                            datatypes=['CP_WHI_NATURAL'], time_clip=True)
        self.assertTrue('Electric_Spectral_Power_Density__C1_CP_WHI_NATURAL' in whi_data)
        self.assertTrue(data_exists('Electric_Spectral_Power_Density__C1_CP_WHI_NATURAL'))

    def test_load_csa_JP_AUX_PMP_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2017-01-01T00:00:00Z', '2017-02-01T00:00:00Z']
        jp_data = load_csa(probes=['C1'],
                           trange=trange,
                           datatypes=['JP_AUX_PMP'], time_clip=True)
        print(jp_data)
        self.assertTrue('L_value__C1_JP_AUX_PMP' in jp_data)
        self.assertTrue(data_exists('L_value__C1_JP_AUX_PMP'))

    def test_load_csa_JP_PSE_data(self):
        self.clean_data()
        config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        trange = ['2017-01-01T00:00:00Z', '2018-01-02T00:00:00Z']
        jp_data = load_csa(probes=['C1'],
                           trange=trange,
                           datatypes=['JP_AUX_PSE'], time_clip=True)
        print(jp_data)
        self.assertTrue('sc_r1_xyz_gse__C1_JP_AUX_PSE' in jp_data)
        self.assertTrue(data_exists('sc_r1_xyz_gse__C1_JP_AUX_PSE'))

if __name__ == '__main__':
    unittest.main(verbosity=2)




import logging
from pyspedas.tplot_tools import store_data, options, set_units, set_coords

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def spd_pgs_moments_tplot(moments, x=None, prefix='', suffix='', coords='DSL', use_mms_sdc_units=False):
    """
    Creates tplot variables from moments dictionaries

    Parameters
    ----------
        moments: dict
            Dictionary containing moments values returned by moments_3d

        x: numpy.ndarray
            The x-axis (time) values

        prefix: str
            Name prefix for the output variables

        suffix: str
            Name suffix for the output variables

        coords:str
            Coordinate system to set for non-field-aligned moments that have coordinates
            Default: 'DSL'

        use_mms_sdc_units: bool
            If True, convert pressure tensor and heat flux values and units to nPa and mW/m^2 respectively,
            for compatibility with MMS SDC products.
            Default: False

    Returns
    -------
    list of str
        List of tplot variables returned
    """

    if x is None:
        logging.error('Error, no x-values specified')
        return None

    if not isinstance(moments, dict):
        logging.error('Error, the "moments" variable must be a hash table containing the moments')
        return None

    units_dict = {
        'density': '1/cm^3',
        'velocity': 'km/s',
        'vthermal': 'km/s',
        'flux': '1/(cm^2-s)',
        'ttens': 'eV',
        't3': 'eV',
        'magt3': 'eV',
        'avgtemp': 'eV',
        'sc_pot': 'V',
        'eflux': 'eV/(cm^2-s)',
        'qflux': 'eV/(cm^2-s)',
        'mftens': 'eV/cm^3',
        'ptens': 'eV/cm^3',
        'symm': 'nT',
        'symm_theta': 'degrees',
        'symm_phi': 'degrees',
        'symm_ang': 'degrees',
        'magf': 'nT',
        }

    conversion_dict = {
        'density': 1.0,
        'velocity': 1.0,
        'vthermal': 1.0,
        'flux': 1.0,
        'ttens': 1.0,
        't3': 1.0,
        'magt3': 1.0,
        'avgtemp': 1.0,
        'sc_pot': 1.0,
        'eflux': 1.0,
        'qflux': 1.0,
        'mftens': 1.0,
        'ptens': 1.0,
        'symm': 1.0,
        'symm_theta': 1.0,
        'symm_phi': 1.0,
        'symm_ang': 1.0,
        'magf': 1.0,
        }

    if use_mms_sdc_units:
        # Convert pressure tensor and heat flux to MMS SDC-compatible units
        units_dict['ptens'] = 'nPa'
        conversion_dict['ptens'] = 0.000160217663
        units_dict['qflux'] = 'mW/m^2'
        conversion_dict['qflux'] = 1.6021765974585869e-12

    fa_coords = ['magf', 'magt3', 'symm']
    non_fa_coords = ['velocity','flux', 't3', 'eflux', 'qflux', 'mftens', 'ptens']


    for key in moments.keys():
        # Some moments (qflux, ptens) may have an optional conversion to different units
        conversion = conversion_dict.get(key)
        store_data(prefix + key + suffix, data={'x': x, 'y': conversion*moments[key]})
        units = units_dict.get(key)
        if units is not None:
            set_units(prefix + key + suffix, units)
            options(prefix+key+suffix, 'ysubtitle', '[' + units + ']')
        if key in fa_coords:
            set_coords(prefix + key + suffix, 'FA')
        elif key in non_fa_coords:
            set_coords(prefix + key + suffix, coords)

    options(prefix + 'velocity' + suffix, 'yrange', [-800, 800])
    # Flux is really spiky, so it's hard to set good default limits...
    options(prefix + 'flux' + suffix, 'yrange', [-1e10, 1e10])

    return [prefix + key + suffix for key in moments.keys()]



from .cammice import cammice
from .ceppad import ceppad
from .mfe import mfe
from .efi import efi
from .pwi import pwi
from .hydra import hydra
from .tide import tide
from .timas import timas
from .uvi import uvi
from .vis import vis
from .pixie import pixie
from .orbit import orbit
from .datasets import datasets



import datetime
import logging
from pyspedas.tplot_tools import get_data, store, tnames
import numpy as np


def tinterpol(names, interp_to, method=None, newname=None, extrapolate=False, suffix=None):
    """
    Interpolate data to times in interp_to.

    Parameters
    ----------
    names : array_like or str
        List of tplot variables to interpolate.
        Allowed wildcards are ? for a single character, * from multiple characters.
    interp_to : array_like or str
        String containing the tplot variable with the time stamps to interpolate to.
    method : str, optional
        Interpolation method. Default is 'linear'.
        Specifies the kind of interpolation as a string ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic', 'previous', 'next')
        where 'zero', 'slinear', 'quadratic' and 'cubic' refer to a spline interpolation of zeroth, first, second or third order;
        'previous' and 'next' simply return the previous or next value of the point) or as an integer specifying the order of the spline interpolator to use.
    newname : str or list of str, optional
        List of new names for tplot variables. If '', then tplot variables are replaced. If not given, then a suffix is applied.
    extrapolate: bool
        If true, extrapolate beyond the start/end times of the interp_to variable.  Default is False (for now)
    suffix : str, optional
        A suffix to apply. Default is '-itrp'.

    Returns
    -------
    None
        This function works in-place or creates new tplot variables depending on the 'newname' parameter.


    Notes
    -----

    Uses xarray interp method to interpolate data to the times in interp_to.
    See: https://docs.xarray.dev/en/latest/generated/xarray.DataArray.interp.html
    Similar to tinterpol.pro in IDL SPEDAS.

    'linear' vs. 'slinear' methods:  Due to a quirk in the implmentation of the scipy interp1d routine (used
    internally by xarray.interp),'linear' interpolation may yield unexpected results under certain conditions.
    In particular, when using 32-bit floating point data, if an output time exactly matches one of the input times, the
    interpolated value may differ slightly (on the order of 1 ULP) from the input value at that timestamp.
    This can cause problems for downstream calculations: for example, if the input is strictly non-negative, but
    contains zero values, the interpolated data may contain small negative values. The 'slinear' method does not appear
    to suffer from this issue, but may be slightly slower.


    Examples
    --------
        >>> import numpy as np
        import pyspedas
        >>> import pyspedas

        >>> # Create some time series data
        >>> times = np.array(['2002-02-03T04:05:00', '2002-02-03T04:05:05', '2002-02-03T04:05:10'], dtype='datetime64')
        >>> data = np.array([0.0, 5.0, 10.0])

        >>> # Store the data as a tplot variable
        >>> pyspedas.store_data('variable1', data={'x': times, 'y': data})

        >>> # Create some new times to interpolate to
        >>> new_times = np.array(['2002-02-03T04:05:07', '2002-02-03T04:05:08', '2002-02-03T04:05:09'], dtype='datetime64')
        >>> new_data = np.array([11.0, 12.0, 13.0])

        >>> # Store the new times in a tplot variable, the new_data will be ignored in this example
        >>> pyspedas.store_data('variable2', data={'x': new_times, 'y': new_data})

        >>> # Interpolate 'variable1' to the new times using the default linear method
        >>> pyspedas.tinterpol(names='variable1', interp_to='variable2')

        >>> # The interpolated data is now stored in 'variable1-itrp'
        >>> interpolated_data = pyspedas.get_data('variable1-itrp')
        >>> print(interpolated_data)
    """
    if not isinstance(names, list):
        names = [names]
    if not isinstance(newname, list):
        newname = [newname]

    old_names = tnames(names)  # expand any wildcards

    if len(old_names) < 1:
        logging.error("tinterpol error: No pytplot names were provided.")
        return

    if suffix is None:
        suffix = "-itrp"

    if method is None:
        method = "linear"

    if (newname is None) or (len(newname) == 1 and newname[0] is None):
        n_names = [s + suffix for s in old_names]
    else:
        n_names = newname

    if isinstance(interp_to, str):
        interp_to_data = get_data(interp_to, dt=True)

        if interp_to_data is None:
            logging.error("Error, tplot variable: " + interp_to + " not found.")
            return

        interp_to_times = interp_to_data[0]
    else:
        interp_to_times = interp_to

    kwargs={}
    if extrapolate:
        kwargs={'fill_value':'extrapolate'}


    for name_idx, name in enumerate(old_names):
        xdata = get_data(name, xarray=True)
        metadata = get_data(name, metadata=True)

        if isinstance(interp_to_times[0], (datetime.datetime, np.datetime64)):
            # Timezone-naive datetime or np.datetime64, use as-is
            interp_to_datetimes = interp_to_times
        elif isinstance(interp_to_times[0], (int, float, np.integer, np.float64)):
            # Assume seconds since Unix epoch, convert to np.datetime64 with nanosecond precision
            if isinstance(interp_to_times, np.ndarray):
                interp_to_datetimes = np.array(
                    interp_to_times * 1e09, dtype="datetime64[ns]"
                )
            else:
                # We need to convert input to a numpy array before scaling to nanoseconds
                interp_to_datetimes = np.array(
                    np.array(interp_to_times) * 1e9, dtype="datetime64[ns]"
                )
        elif isinstance(interp_to_times[0], str):
            # Interpret strings as timestamps, convert to np.datetime64 with nanosecond precision
            interp_to_datetimes = np.array(interp_to_times, dtype="datetime64[ns]")
        else:
            # Give up for any other type
            logging.error(
                "tinterpol: Unable to convert type %s to timestamp.",
                type(interp_to_times[0]),
            )
            return

        xdata_interpolated = xdata.interp({"time": interp_to_datetimes}, method=method, kwargs=kwargs)

        if "spec_bins" in xdata.coords:
            store(
                n_names[name_idx],
                data={
                    "x": interp_to_times,
                    "y": xdata_interpolated.values,
                    "v": xdata_interpolated.coords["spec_bins"].values,
                },
                metadata=metadata,
            )
        else:
            store(
                n_names[name_idx],
                data={"x": interp_to_times, "y": xdata_interpolated.values},
                metadata=metadata,
            )

        logging.info("tinterpol (" + method + ") was applied to: " + n_names[name_idx])



from .load import load

# This routine was moved out of __init__.py.  Please see that file for previous revision history.

def fspc(trange=['2013-01-29','2013-01-30'],
        probe='1A',
        downloadonly=False,
        prefix='',
        suffix='',
        no_update=False,
        time_clip=False,
        force_download=False):
    """
    This function load data from BARREL Fast Spectra (4 channels record at 20Hz)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2013-01-29','2013-01-30']

        probe: str
            Options::

                '1a'...'1v'
                '2a'...'2y'
                '3a'...'3g'
                '4a'...'4f'
                '5a'
                '6a'
                '7a'

            Default: '1A'

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables.
            Default: False

        prefix: str
            This string will be prepended to the tplot variable names created. Default: ''

        suffix: str
            This string will be appended to the tplot variable names created.  Default: ''

        no_update: bool
            If set, only load data from your local cache
            Default: False
        
        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Returns
    ----------
        List of tplot variables created.

    Example:
    ----------
        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> pyspedas.projects.barrel.fspc(trange=['2013-01-17', '2013-01-19'], probe='1D')
        >>> tplot('brl1D_FSPC1')
    """

    return load(datatype='fspc',trange=trange, probe=probe,  prefix=prefix, suffix=suffix, downloadonly=downloadonly, time_clip=time_clip, no_update=no_update, force_download=force_download)





import pyspedas
from pyspedas.tplot_tools import time_double
from pyspedas.tplot_tools import xlim
import logging
import numpy as np


def timespan(t1=None, dt=None, units="days", keyword=None, reset=False):
    """
    This function will set the time range for all time series plots. This is a wrapper for the function "xlim" to
    better handle time axes.

    Parameters
    ----------
    t1 : float or str
        The time to start all time series plots. Can be given in seconds since epoch, or as a string
        in the format "YYYY-MM-DD HH:MM:SS".
    dt : float
        The time duration of the plots. Default is number of days.
    units : str, optional
        Sets the units of the "dt" variable. Days, hours, minutes, and seconds and reasonable abbreviations are all accepted. Default is 'days'.
    keyword : str, optional
        Synonym for units keyword
    reset: bool, optional
        If True, clears all previously set time ranges.

    Returns
    -------
    None

    Examples
    --------
    >>> # Set the timespan to be 2017-07-17 00:00:00 plus 1 day
    >>> import pyspedas
    >>> pyspedas.timespan(1500249600, 1)

    >>> # The same as above, but using different inputs
    >>> pyspedas.timespan("2017-07-17 00:00:00", 24, units='hours')
    """

    if reset is True:
        if 'x_range_full' in pyspedas.tplot_tools.tplot_opt_glob.keys():
            del pyspedas.tplot_tools.tplot_opt_glob['x_range_full']
        if 'x_range_last' in pyspedas.tplot_tools.tplot_opt_glob.keys():
            del pyspedas.tplot_tools.tplot_opt_glob['x_range_last']
        if 'x_range' in pyspedas.tplot_tools.tplot_opt_glob.keys():
            del pyspedas.tplot_tools.tplot_opt_glob['x_range']
        if 'xfull' in pyspedas.tplot_tools.lim_info.keys():
            del pyspedas.tplot_tools.lim_info['xfull']
        if 'xlast' in pyspedas.tplot_tools.lim_info.keys():
            del pyspedas.tplot_tools.lim_info['xlast']
        return
    elif isinstance(t1, (list, tuple, np.ndarray)):
        if len(t1) != 2:
            logging.warning('timespan: If input is a list or array, it must have 2 elements')
            return None
        else:
            t1 = pyspedas.tplot_tools.time_double(t1)
            xlim(t1[0], t1[1])
            return
    elif isinstance(t1, str):
        t1 = pyspedas.tplot_tools.time_double(t1)

    if keyword is not None:
        units = keyword

    if dt is None or not isinstance(dt, (int, float)):
        logging.warning('timespan: Invalid value for dt: ', dt)
    if units.lower() in ["days", 'day', 'd']:
        dt *= 86400
    elif units.lower() in ["hours", 'hour', 'hr', 'hrs', 'h']:
        dt *= 3600
    elif units.lower() in ["minutes", 'minute', 'min', 'mins', 'm']:
        dt *= 60
    elif units.lower() in ["seconds", 'second', 'sec', 'secs', 's']:
        dt *= 1
    else:
        logging.warning(
            "Invalid 'units' option %s.\nEnum(None, 'days', 'day', 'd', 'hours', 'hour', 'hr', 'hrs', 'h', 'minutes', 'minute', 'min', 'mins', 'm', 'seconds', 'second', 'secs', 'sec', 's')",
            keyword,
        )

    if not isinstance(t1, (int, float, complex)):
        t1 = pyspedas.tplot_tools.time_double(t1)
    t2 = t1 + dt
    xlim(t1, t2)

    return



from pyspedas import find_datasets

# This routine was originally in lanl/__init__.py.
def datasets(instrument=None, label=True):
    return find_datasets(mission="LANL", instrument=instrument, label=label)



import numpy as np
from pyspedas import clip, get_data, options, ylim

from ..load import load
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional

def mgf(
    trange: List[str] = ['2017-03-27', '2017-03-28'],
    datatype: str = '8sec',
    level: str = 'l2',
    suffix: str = '',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    coord: str = 'dsi',
    version: Optional[str] = None,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the MGF experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-03-27', '2017-03-28']

        datatype: str
            Data type; Valid options: '128hz', '256hz', '64hz', '8sec'

        level: str
            Data level; Valid options: 'l2'  Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix.  Default: None

        get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot. Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        ror: bool
            If set, print PI info and rules of the road
            Default: True

        coord: str
            Valid values: "sm", "dsi", "gse", "gsm", "sgi"
            Default: 'dsi'

        version: str
            Set this value to specify the version of cdf files (such as "v03.03", "v03.04", ...)

        uname: str
            User name. Default: None

        passwd: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> mgf_vars = pyspedas.projects.erg.mgf(trange=['2017-03-27', '2017-03-28'])
    >>> tplot('erg_mgf_l2_mag_8sec_sm')

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    if datatype == '8s' or datatype == '8':
        datatype = '8sec'
    elif datatype == '64':
        datatype = '64hz'
    elif datatype == '128':
        datatype = '128hz'
    elif datatype == '256':
        datatype = '256hz'

    prefix = 'erg_mgf_'+level+'_'
    if datatype == '8sec':
        file_res = 3600. * 24
        pathformat = 'satellite/erg/mgf/'+level+'/'+datatype + \
            '/%Y/%m/erg_mgf_'+level+'_'+datatype+'_%Y%m%d_'
    else:
        file_res = 3600.
        pathformat = 'satellite/erg/mgf/'+level+'/'+datatype + \
            '/%Y/%m/erg_mgf_'+level+'_'+datatype+'_' + coord + '_%Y%m%d%H_'

    if version is None:
        pathformat += 'v??.??.cdf'
    else:
        pathformat += version + '.cdf'

    loaded_data = load(pathformat=pathformat, file_res=file_res, trange=trange, level=level, datatype=datatype, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)

    if (loaded_data is None) or (loaded_data == []):
        return loaded_data

    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)

            # --- print PI info and rules of the road
            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG MGF')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: ", gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                'RoR of MGF L2: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Mgf')
            print('Contact: erg_mgf_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if datatype == '8sec':

        # remove -1.0e+30

        clip(prefix + 'mag_'+datatype+'_dsi'+suffix, -1e+6, 1e6)
        clip(prefix + 'mag_'+datatype+'_gse'+suffix, -1e+6, 1e6)
        clip(prefix + 'mag_'+datatype+'_gsm'+suffix, -1e+6, 1e6)
        clip(prefix + 'mag_'+datatype+'_sm'+suffix, -1e+6, 1e6)

        clip(prefix + 'magt_'+datatype+suffix, -1e+6, 1e6)

        clip(prefix + 'rmsd_'+datatype+'_dsi'+suffix, -1e+6, +1e+6)
        clip(prefix + 'rmsd_'+datatype+'_gse'+suffix, -1e+6, +1e+6)
        clip(prefix + 'rmsd_'+datatype+'_gsm'+suffix, -1e+6, +1e+6)
        clip(prefix + 'rmsd_'+datatype+'_sm'+suffix, -1e+6, +1e+6)

        clip(prefix + 'rmsd_'+datatype+suffix, 0., 80.)

        clip(prefix + 'dyn_rng_'+datatype+suffix, -120., +1e+6)

        clip(prefix + 'igrf_'+datatype+'_dsi'+suffix, -1e+6, +1e+6)
        clip(prefix + 'igrf_'+datatype+'_gse'+suffix, -1e+6, +1e+6)
        clip(prefix + 'igrf_'+datatype+'_gsm'+suffix, -1e+6, +1e+6)
        clip(prefix + 'igrf_'+datatype+'_sm'+suffix, -1e+6, +1e+6)

        # set yrange
        _, bdata = get_data(prefix + 'mag_'+datatype+'_dsi'+suffix)
        ylim(prefix + 'mag_'+datatype+'_dsi'+suffix,
             np.nanmin(bdata), np.nanmax(bdata))
        _, bdata = get_data(prefix + 'mag_'+datatype+'_gse'+suffix)
        ylim(prefix + 'mag_'+datatype+'_gse'+suffix,
             np.nanmin(bdata), np.nanmax(bdata))
        _, bdata = get_data(prefix + 'mag_'+datatype+'_gsm'+suffix)
        ylim(prefix + 'mag_'+datatype+'_gsm'+suffix,
             np.nanmin(bdata), np.nanmax(bdata))
        _, bdata = get_data(prefix + 'mag_'+datatype+'_sm'+suffix)
        ylim(prefix + 'mag_'+datatype+'_sm'+suffix,
             np.nanmin(bdata), np.nanmax(bdata))

        _, bdata = get_data(prefix + 'magt_'+datatype+suffix)
        ylim(prefix + 'magt_'+datatype+suffix,
             np.nanmin(bdata), np.nanmax(bdata))

        _, bdata = get_data(prefix + 'rmsd_'+datatype+suffix,)
        ylim(prefix + 'rmsd_'+datatype+suffix,
             np.nanmin(bdata), np.nanmax(bdata))

        _, bdata = get_data(prefix + 'rmsd_'+datatype+'_dsi'+suffix)
        ylim(prefix + 'rmsd_'+datatype+'_dsi'+suffix,
             np.nanmin(bdata), np.nanmax(bdata))
        _, bdata = get_data(prefix + 'rmsd_'+datatype+'_gse'+suffix)
        ylim(prefix + 'rmsd_'+datatype+'_gse'+suffix,
             np.nanmin(bdata), np.nanmax(bdata))
        _, bdata = get_data(prefix + 'rmsd_'+datatype+'_gsm'+suffix)
        ylim(prefix + 'rmsd_'+datatype+'_gsm'+suffix,
             np.nanmin(bdata), np.nanmax(bdata))
        _, bdata = get_data(prefix + 'rmsd_'+datatype+'_sm'+suffix)
        ylim(prefix + 'rmsd_'+datatype+'_sm'+suffix,
             np.nanmin(bdata), np.nanmax(bdata))

        _, bdata = get_data(prefix + 'rmsd_'+datatype+suffix)
        ylim(prefix + 'rmsd_'+datatype+suffix,
             np.nanmin(bdata), np.nanmax(bdata))

        _, bdata = get_data(prefix + 'quality_'+datatype+suffix)
        ylim(prefix + 'quality_'+datatype+suffix,
             np.nanmin(bdata), np.nanmax(bdata))
        _, bdata = get_data(prefix + 'quality_'+datatype+'_gc'+suffix)
        ylim(prefix + 'quality_'+datatype+'_gc' +
             suffix, np.nanmin(bdata), np.nanmax(bdata))

        # set labels
        options(prefix + 'mag_'+datatype+'_dsi'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])
        options(prefix + 'mag_'+datatype+'_gse'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])
        options(prefix + 'mag_'+datatype+'_gsm'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])
        options(prefix + 'mag_'+datatype+'_sm'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])

        options(prefix + 'rmsd_'+datatype+'_dsi'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])
        options(prefix + 'rmsd_'+datatype+'_gse'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])
        options(prefix + 'rmsd_'+datatype+'_gsm'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])
        options(prefix + 'rmsd_'+datatype+'_sm'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])

        options(prefix + 'igrf_'+datatype+'_dsi'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])
        options(prefix + 'igrf_'+datatype+'_gse'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])
        options(prefix + 'igrf_'+datatype+'_gsm'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])
        options(prefix + 'igrf_'+datatype+'_sm'+suffix,
                'legend_names', ['Bx', 'By', 'Bz'])

        # set color of the labels
        options(prefix + 'mag_'+datatype+'_dsi' +
                suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'mag_'+datatype+'_gse' +
                suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'mag_'+datatype+'_gsm' +
                suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'mag_'+datatype+'_sm' +
                suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'rmsd_'+datatype+'_dsi' +
                suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'rmsd_'+datatype+'_gse' +
                suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'rmsd_'+datatype+'_gsm' +
                suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'rmsd_'+datatype+'_sm' +
                suffix, 'Color', ['b', 'g', 'r'])

        options(prefix + 'quality_'+datatype+suffix, 'Color', ['r', 'g', 'b'])

        options(prefix + 'igrf_'+datatype+'_dsi' +
                suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'igrf_'+datatype+'_gse' +
                suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'igrf_'+datatype+'_gsm' +
                suffix, 'Color', ['b', 'g', 'r'])
        options(prefix + 'igrf_'+datatype+'_sm' +
                suffix, 'Color', ['b', 'g', 'r'])
    else:
        # remove -1.0e+30
        clip(prefix + 'mag_'+datatype+'_' + coord + suffix, -1e+6, 1e6)
        # set yrange
        _, bdata = get_data(prefix + 'mag_'+datatype+'_' + coord + suffix)
        ylim(prefix + 'mag_'+datatype+'_' + coord +
             suffix, np.nanmin(bdata), np.nanmax(bdata))
        # set labels
        options(prefix + 'mag_'+datatype+'_' + coord +
                suffix, 'legend_names', ['Bx', 'By', 'Bz'])
        # set color of the labels
        options(prefix + 'mag_'+datatype+'_' + coord +
                suffix, 'Color', ['b', 'g', 'r'])
    return loaded_data
